#!/usr/bin/env python3

# This program was written by Peter Kristoffersen
# (petek@stud.ntnu.no) as part of his master's thesis.
# It is based on previous programming work by
# Maryam Gholami Mayani and Turid Worren Reenaas which
# is in turn based on theoretical work by Rune Strandberg
#
# Although the program was written for python 3.x,
# it will run with minor changes in python 2, specifically
# changing the names of some unicode variable identifiers.

import numpy as np
import scipy.integrate
import time
import datetime as dt


class C:  # Constants
    q = 1.602176487e-19  # Elementary charge [C]
    k = 8.6173324e-5  # Boltzmann's constant [eV/K]
    h = 4.135667516e-15  # Planck constant [eV s]
    c = 2.99792e8  # Speed of light [m/s]
    pi = np.pi
    k_strange = 1.38065E-19  # Boltzmann constant in weird units [kg cm^2/(s^2 K)]
    h_strange = 6.62607E-30  # Planck constant [kg cm^2/s]


class Error:
    err_negative_current = 2 ** 0  # 1
    err_n_ib_negative = 2 ** 1  # 2
    err_n_ib_too_large = 2 ** 2  # 4
    err_mu_vi_too_large = 2 ** 3  # 8
    err_mu_ic_too_large = 2 ** 4  # 16


class Result:
    def __init__(self):
        for k in ("ib_net_flux", "E_0",
                  "G_vc", "G_vi", "G_ic",
                  "R_vc", "R_vi", "R_ic",
                  "SRH_vc", "SRH_vi", "SRH_ic",
                  "auger_vc", "auger_vi", "auger_ic",
                  "n", "p", "n_t", "n_ib",
                  "F_v", "F_c", "F_t", "F_ib",
                  "V", "V_shift", "P", "η", "J", "errors"):
            setattr(self, k, None)


class Parameters:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

        def m_from_N(N, T):
            """
            Calculates effective mass from carrier concentration
            :param N: carrier concentration in cm^-3
            :param T: cell temperature in K
            return effective mass in kg
            """
            return (N / 2) ** (2 / 3) * C.h_strange ** 2 / (2 * C.pi * C.k_strange * T)

        if not hasattr(self, "m_e"):
            self.m_e = m_from_N(self.N_c, self.T)
        if not hasattr(self, "p_e"):
            self.m_p = m_from_N(self.N_v, self.T)
        if not hasattr(self, "E_vi"):
            self.E_vi = self.E_g - self.E_ic
        if not hasattr(self, "E_ic"):
            self.E_ic = self.E_g - self.E_vi
        if hasattr(self, "auger_C"):
            if not hasattr(self, "auger_C_T1"):
                self.auger_C_T1 = self.auger_C
            if not hasattr(self, "auger_C_T2"):
                self.auger_C_T2 = self.auger_C
            if not hasattr(self, "auger_C_T3"):
                self.auger_C_T3 = self.auger_C
            if not hasattr(self, "auger_C_T4"):
                self.auger_C_T4 = self.auger_C
            if not hasattr(self, "auger_C_B1"):
                self.auger_C_B1 = self.auger_C
            if not hasattr(self, "auger_C_B2"):
                self.auger_C_B2 = self.auger_C

    # Default simulation parameters, can be overridden
    T = 300  # Room temperature [K]
    T_s = 6000  # Temperature of sun [K]
    sun_x = 1  # Light concentration [1]
    max_integration = 18  # [eV], sets the upper limit for black body integration. Set to np.inf only if needed

    # Default program parameters, can be overwritten
    N_iter = 50  # Search iterations for equilibrium values
    N_iter2 = 20  # Search iterations for detailed balance
    V_num = 100  # Number of voltages to calculate
    calculate_auger = True
    calculate_SRH = True

    @classmethod
    def from_import(cls, import_name):
        import importlib.machinery
        loader = importlib.machinery.SourceFileLoader('input_file', import_name)
        s = loader.load_module()
        init_dict = dict((k, v) for k, v in s.__dict__.items() if not k.startswith("__"))
        return cls(**init_dict)

    @classmethod
    def from_dict(cls, dictionary):
        return cls(**dictionary)


class Experiment:  # Base class for experiments
    def __init__(self, parameters):
        self.parameters = parameters
        self.results = Result()

    def calculate_SRH(self, N0, N1, Et, n, n0, p, p0, Fn, Fp):
        P = self.parameters
        B = 1 / (C.k * P.T)  # eV^-1
        v_n = np.sqrt(3 * C.k_strange * P.T / P.m_e)  # cm/s
        v_p = np.sqrt(3 * C.k_strange * P.T / P.m_p)  # cm/s
        τ_n = 1 / (v_n * P.σ_n * P.N_t)  # seconds
        τ_p = 1 / (v_p * P.σ_p * P.N_t)  # seconds
        nt = N1 * np.exp((Et - Fn) * B)  # cm^-3
        pt = N0 * np.exp((Fp - Et) * B)  # cm^-3
        R = (n * p - n0 * p0) / (τ_p * (n + nt) + τ_n * (p + pt))  # cm^-3 s^-1
        return R * P.w  # cm^-2 s^-1

    def calculate(self):
        ## Set up shortcuts
        P = self.parameters
        R = self.results
        B = 1 / (C.k * P.T)
        B_s = 1 / (C.k * P.T_s)

        ## PART 1:
        ## Calculate equilibrium filling and energies using poisson eq.
        ##
        ## We use binary search of the poisson equation to find poisson(E_0)=0
        ## In most (all) cases, poisson(E_0) will be monotonically increasing.
        ## We rely on this fact.

        def poisson(E_0):
            p0 = P.N_v / (np.exp(E_0 * B))  # holes in VB with E_i
            n0 = P.N_c / (np.exp((P.E_g - E_0) * B))  # electrons in CB with E_i
            n0_ib = P.N_ib / (1 + np.exp((P.E_vi - E_0) * B))  # ib states filled
            n0_t = P.N_t / (1 + np.exp((P.E_t - E_0) * B))  # trap states filled
            charge = n0 + n0_ib + n0_t + P.N_a - p0 - P.N_d
            return charge, p0, n0, n0_ib, n0_t

        # This method can fail if N_v is very large compared to the other N
        # Then, poisson(E_i) does not monotonically increase
        # Then there might also be two zeros.... I think
        b_min, b_max = 0, P.E_g
        R.E_0 = (b_max + b_min) / 2
        for _ in range(P.N_iter):  # This is the binary search
            if poisson(R.E_0)[0] > 0:
                b_max = R.E_0
            else:
                b_min = R.E_0
            R.E_0 = (b_max + b_min) / 2
        p0, n0, n0_ib, n0_t = poisson(R.E_0)[1:]
        p0_ib = P.N_ib - n0_ib

        ##
        ## PART 2:
        ## We solve the continuity equation.
        ## For steady state, the net charge flow into the IB must be 0.
        ## For each fermi level split, V, we find the Vshift, the distance
        ## from the VB edge to the VB fermi level that fixes net IB current to 0.
        R.V = np.linspace(0, P.E_g, P.V_num + 2)[1:-1]  # Skip first and last

        # Set up constants and values that is used in the calculation:
        θ_c = np.arcsin(1 / P.nr)  # Critical angle for internal reflection
        cons_frac = P.sun_x / 46050  # Concentration fraction
        # Integrals for detailed balance current production
        s = 1E-4 * (2 * P.nr ** 2) / (C.h ** 3 * C.c ** 2)  # Prefactor for integral [cm^-2 eV^-3 s^-1]

        def DB_integral(E_low, E_high, µ, B):  # Black body flux integral
            """Calculates the flux of photons with energy in the range
               E_low to E_high from a black body with chemical potential
               µ and temperature B = 1/kT
               µ is an array of chemical potentials."""

            def f(x, µ):
                return x ** 2 / (np.exp((x - µ) * B) - 1)

            return s * np.array([scipy.integrate.quad(f, E_low, E_high, (m,))[0] for m in µ])  # [cm^-2 s^-1]

        def DB_integral2(critical_angle, α):  # Internal reflection integral
            """Calculates the integral for internal reflection.
               :param critical_angle: Critical angle
               :param α: array (!) of absorption coefficients
               Returns an array (!) of the results"""

            def f(x, a):
                cs = np.cos(x)
                sn = np.sin(x)
                return (1 - np.exp(-2 * P.w * a / cs)) * cs * sn

            return 2 * np.pi * np.array([scipy.integrate.quad(f, 0, critical_angle, (ai,))[0] for ai in α])

        if P.E_vi > P.E_ic:
            vi_flux_s = DB_integral(P.E_vi, P.E_g, np.zeros(1), B_s)[0]  # Solar photons in vi energy range
            ic_flux_s = DB_integral(P.E_ic, P.E_vi, np.zeros(1), B_s)[0]  # Solar photons in ic energy range
        else:  # Eic is larger than Evi
            vi_flux_s = DB_integral(P.E_vi, P.E_ic, np.zeros(1), B_s)[0]  # Solar photons in vi energy range
            ic_flux_s = DB_integral(P.E_ic, P.E_g, np.zeros(1), B_s)[0]  # Solar photons in ic energy range
        vc_flux_s = DB_integral(P.E_g, P.max_integration, np.zeros(1), B_s)[0]

        def recombination(R):
            """Given an array of Vshift,
               calculates total flux of carriers into IB.
               Should be 0 for steady state.
               Calculates for all values of (V,Vshift)."""
            lenV = len(R.V)
            R.errors = np.zeros(lenV, dtype=np.int32)  # bitmask of errors
            # Calculate fermi levels
            R.F_v = R.Vshift
            R.F_c = R.Vshift + R.V
            R.F_t = R.F_v  # Simplified
            # Calculate carrier concentrations
            R.p = P.N_v / (np.exp(R.F_v * B) + 1)
            R.n = P.N_c / (np.exp((P.E_g - R.F_c) * B) + 1)
            R.n_t = P.N_t / (np.exp((P.E_t - R.F_t) * B) + 1)
            R.n_ib = R.p + P.N_d - R.n - R.n_t - P.N_a  # Since charge is conserved, rho_f = 0
            R.errors += (R.n_ib < 0) * Error.err_n_ib_negative  # Check that 0<n_ib<N_ib
            R.errors += (R.n_ib > P.N_ib) * Error.err_n_ib_too_large
            p_ib = P.N_ib - R.n_ib
            R.F_ib = P.E_vi - np.log(P.N_ib / R.n_ib - 1) / B
            # Recombination and generation
            µ_vi = R.F_ib - R.F_v
            µ_ic = R.F_c - R.F_ib
            R.errors += (µ_vi > P.E_vi) * Error.err_mu_vi_too_large
            R.errors += (µ_ic > P.E_ic) * Error.err_mu_ic_too_large
            vi_flux_a = np.zeros(lenV)
            ic_flux_a = np.zeros(lenV)
            int2_vi = np.zeros(lenV)
            int2_ic = np.zeros(lenV)
            # Absorption coefficients
            α_vi = P.σ_vi * p_ib
            α_ic = P.σ_ic * R.n_ib
            int2_vi[R.errors == 0] = DB_integral2(θ_c, α_vi[R.errors == 0])
            int2_ic[R.errors == 0] = DB_integral2(θ_c, α_ic[R.errors == 0])
            # Solar Generation
            R.G_vi = cons_frac * vi_flux_s * int2_vi
            R.G_ic = cons_frac * ic_flux_s * int2_ic
            if P.E_vi > P.E_ic:
                # Ambient photons in vi energy range
                vi_flux_a[R.errors == 0] = DB_integral(P.E_vi, P.E_g, µ_vi[R.errors == 0], B)
                # Ambient photons in ic energy range
                ic_flux_a[R.errors == 0] = DB_integral(P.E_ic, P.E_vi, µ_ic[R.errors == 0], B)
            else:
                # Ambient photons in vi energy range
                vi_flux_a[R.errors == 0] = DB_integral(P.E_vi, P.E_ic, µ_vi[R.errors == 0], B)
                # Ambient photons in ic energy range
                ic_flux_a[R.errors == 0] = DB_integral(P.E_ic, P.E_g, µ_ic[R.errors == 0], B)
            R.R_vi = vi_flux_a * int2_vi
            R.R_ic = ic_flux_a * int2_ic
            # Calculate SRH recombination
            R.SRH_vi = np.zeros(lenV)
            R.SRH_ic = np.zeros(lenV)
            if hasattr(P, 'calculate_SRH') and P.calculate_SRH is True:
                if P.E_t < P.E_vi:  # SRH from IB to VB via trap
                    R.SRH_vi = self.calculate_SRH(P.N_v, P.N_ib, P.E_t, R.n_ib, n0_ib, R.p, p0, R.F_ib, R.F_v)
                    R.SRH_ic = np.zeros(lenV)
                elif P.E_t > P.E_vi:  # SRH from CB to IB via trap
                    R.SRH_vi = np.zeros(lenV)
                    R.SRH_ic = self.calculate_SRH(P.N_ib, P.N_c, P.E_t, R.n, n0, p_ib, p0_ib, R.F_c, R.F_ib)

            # Calculate Auger recombination
            if hasattr(P, 'calculate_auger') and P.calculate_auger is False:
                R.auger_ic = np.zeros(lenV)
                R.auger_vi = np.zeros(lenV)
            else:
                R.auger_ic_e = P.auger_C_T1 * R.n * R.n * p_ib * P.w
                R.auger_ic_h = P.auger_C_T2 * R.n * R.p * p_ib * P.w
                R.auger_ic = R.auger_ic_e + R.auger_ic_h
                R.auger_vi_e = P.auger_C_T3 * R.p * R.n * R.n_ib * P.w
                R.auger_vi_h = P.auger_C_T4 * R.p * R.p * R.n_ib * P.w
                R.auger_vi = R.auger_vi_e + R.auger_vi_h
            R.auger_vi *= R.errors == 0
            R.auger_ic *= R.errors == 0
            net_auger = R.auger_ic - R.auger_vi  # Net carrier flow into IB
            net_SRH = R.SRH_ic - R.SRH_vi
            R.ib_net_flux = (R.G_vi - R.G_ic +
                             R.R_ic - R.R_vi +
                             net_auger + net_SRH)  # Net carrier flow into IB

        def divisions(a, b):
            l = [a, b]
            i = 1
            while True:
                new = [n for n in np.linspace(a, b, 2 ** i + 1) if n not in l]
                yield from new
                l.extend(new)
                i += 1

        trials = dict()
        upper_bound = np.zeros(P.V_num, dtype="bool")
        lower_bound = np.zeros(P.V_num, dtype="bool")

        b_min, b_max = (np.zeros(P.V_num),
                        P.E_g - R.V)

        R.Vshift = (b_max + b_min) / 2
        indexes = np.arange(P.V_num)

        for counter in range(P.N_iter2):  # This is the binary search
            recombination(R)
            errors = R.errors != 0
            too_high_n_ib = (R.errors & Error.err_n_ib_too_large) != 0
            too_low_n_ib = (R.errors & Error.err_n_ib_negative) != 0
            negative = R.ib_net_flux < 0
            set_lower = too_high_n_ib | (negative & ~errors)
            set_upper = too_low_n_ib | (~negative & ~errors)
            upper_bound |= set_upper
            b_max[set_upper] = R.Vshift[set_upper]
            lower_bound |= set_lower
            b_min[set_lower] = R.Vshift[set_lower]
            updated = set_upper | set_lower

            indeterminate = errors & ~too_high_n_ib & ~too_low_n_ib

            # If no bounds are known, or if the current Vshift doesn't give enough info:
            no_bounds = (~upper_bound & ~lower_bound & ~updated) | indeterminate
            check_index = indexes[no_bounds]
            for k in list(trials.keys()):
                if k not in check_index:
                    del trials[k]
            for j in check_index:
                if j not in trials:
                    trials[j] = divisions(b_min[j], b_max[j])
                R.Vshift[j] = next(trials[j])

            # Case no upper bound
            no_upper = ~upper_bound & lower_bound & ~updated
            b_max[no_upper] = (b_max[no_upper] + b_min[no_upper]) / 2
            # Case no lower bound
            no_lower = ~lower_bound & upper_bound & ~updated
            b_min[no_lower] = (b_max[no_lower] + b_min[no_lower]) / 2
            # Case both bounds

            R.Vshift[~no_bounds] = ((b_max + b_min) / 2)[~no_bounds]

        del trials

        recombination(R)
        ## PART 3:
        ## Now that we have solved the continuity equation, we calculate
        ## the efficiency.

        # Generation vc
        int2_vc = DB_integral2(θ_c, np.ones(1) * P.α_vc)
        R.G_vc = np.ones(P.V_num) * cons_frac * vc_flux_s * int2_vc[0]

        # Recombination vc
        vc_flux_a = DB_integral(P.E_g, P.max_integration, R.V, B)  # Time consuming
        R.R_vc = vc_flux_a * int2_vc

        # SRH vc
        if hasattr(P, 'calculate_SRH') and P.calculate_SRH is True:
            R.SRH_vc = self.calculate_SRH(P.N_v, P.N_c, P.E_t, R.n, n0, R.p, p0, R.F_c, R.F_v)
        else:
            R.SRH_vc = np.zeros(P.V_num)

        # Calculate Auger recombination
        if hasattr(P, 'calculate_auger') and P.calculate_auger is True:
            R.auger_vc_e = P.auger_C_B1 * R.n * R.n * R.p * P.w
            R.auger_vc_h = P.auger_C_B2 * R.n * R.p * R.p * P.w
            R.auger_vc = R.auger_vc_h + R.auger_vc_e
        else:
            R.auger_vc = np.zeros(P.V_num)

        # net losses from CB to VB
        # Since flux into IB = flux out of IB at steady state,
        # only include loss mechanisms from IB->VB
        nonradiative_losses = R.SRH_vc + R.SRH_vi + R.auger_vc + R.auger_vi

        R.J = C.q * (R.G_vc - R.R_vc +
                     R.G_vi - R.R_vi -
                     nonradiative_losses)
        R.errors += (R.J < 0) * Error.err_negative_current

        R.P = R.J * R.V  # W/cm^2

        # efficiency
        A33 = lambda E: E ** 3 / (np.exp(E * B_s) - 1)
        B33 = scipy.integrate.quad(A33, 0, P.max_integration)[0]
        P_sun = cons_frac * ((2 * C.pi) / (C.h ** 3 * C.c ** 2)) * B33  # Sun effect [eV/(m^2 s)]
        P_sun_SI = P_sun * 1.602177E-19 * 1E-4  # Convert to J/(cm^2 s) (1596 W/m^2 for default values)
        R.η = R.P / P_sun_SI

        return "Done"


### Code below is for parsing command line arguments

def input_to_dicts(param_file):
    """Takes an input file and outputs a numpy matrix with a column
       for each variable in the input file and calculated value"""
    import itertools
    from collections import defaultdict
    import importlib.machinery
    import sys
    import copy

    loader = importlib.machinery.SourceFileLoader('input_file', param_file)
    loader.__cached__ = None
    inp = loader.load_module()
    imp_dict = dict((k, v) for k, v in inp.__dict__.items()
                    if isinstance(v, (int, float, list, tuple, str, bool, dict)) and not k.startswith("__"))
    common_input = dict((k, v) for k, v in imp_dict.items() if not hasattr(v, '__iter__'))  
    # All parameters that are not iterable are common to all calculations
    multi_input = dict((k, sorted(v)) for k, v in imp_dict.items() if
                       k not in common_input)  # All parameters that vary for each experiment
    multi_keys = multi_input.keys()
    multi_combs = list(
        itertools.product(*[multi_input[k] for k in multi_keys]))  # Generate all combinations of parameters
    multi_dict = [dict(zip(multi_keys, values)) for values in multi_combs]
    input_columns = defaultdict(list)
    output_columns = defaultdict(list)
    input_dict = copy.deepcopy(multi_dict)
    start_time = time.time()
    for counter, input_variables in enumerate(input_dict):
        input_variables.update(common_input)
        parameters = Parameters.from_dict(input_variables)
        ex = Experiment(parameters)
        time_now = time.time()
        time_elapsed = int(time_now - start_time)
        if counter == 0:
            time_total = 0
        else:
            time_total = (time_elapsed / counter) * len(input_dict)
        time_left = int(time_total - time_elapsed)
        print("Calculating case {}/{} of {}. (Time elapsed: {}. Time left: {})".format(counter + 1,
                                                                                       len(input_dict),
                                                                                       param_file,
                                                                                       dt.timedelta(
                                                                                           seconds=time_elapsed),
                                                                                       dt.timedelta(seconds=time_left)))
        ex.calculate()
        no_outputs = len(ex.results.V)
        for k, i in [(k, i) for k, i in ex.results.__dict__.items() if
                     isinstance(i, np.ndarray) and len(i) == no_outputs]:
            output_columns[k].extend(i)
        for k, i in input_variables.items():
            input_columns[k].extend([i] * no_outputs)
    return input_columns, output_columns, common_input, multi_dict


def main(arguments=None):
    import sys
    import os
    if arguments:
        args = parser.parse_args(arguments.split())
    else:
        args = parser.parse_args()
    if args.format is None:
        args.format = ['txt']
    for input_file in args.input_file:
        if not os.path.isfile(input_file):
            raise Exception("File {} does not exist".format(input_file))
    for input_file in args.input_file:
        print("Processing file {}.".format(input_file))
        inp_folder = os.path.split(input_file)[0]
        input_columns, output_columns, common, multi = input_to_dicts(input_file)
        [writers.writers[writer](input_file, input_columns,
                                 output_columns, common, multi) for writer in writers.writers if writer in args.format]


if __name__ == "__main__":
    import argparse
    import writers

    parser = argparse.ArgumentParser(description="Calculates conditions in a semiconductor under illumination.")
    parser.add_argument('input_file', nargs="+")
    parser.add_argument('-f', '--format', required=False, choices=writers.writers.keys(), nargs="*",
                        help="Specify the format of the output, if not specified, will output .txt file.")
    main()
